
# Please use this model for your research but we ask that you reference this paper 'Virtual Design of 3D printed bone tissue engineered scaffold shape using mechanobiological modelling: relationship of scaffold pore architecture to bone tissue formation'.
# # Bone Regeneration Model 

# In[ ]:


import numpy as np
import pandas as pd
import matplotlib.pyplot  as plt
from itertools import product
import math
import random
from scipy.spatial import ConvexHull, QhullError
import os
import subprocess


# 

# ## 1.Read Job File

# In[ ]:


# read job inp file

with open('Job-1.inp') as f:
    lines = f.readlines()
    
# Check for Parts
parts_index = [i for i, item in enumerate(lines) if '*Instance' in item]
if (len(parts_index) != 0):
    print('Part(s) found')
    for i in parts_index:
        print (lines[i])
else:
    print('No Partfound')

# Traget Part is Callus, for now assuming 1st part is the Traget part.
lines_callus = lines[parts_index[1]:parts_index[2]]


# In[ ]:


lines_callus


# ### 1.1 Read Element

# In[ ]:


## Nodes

# get only nodes
nodes_list = lines_callus[lines_callus.index('*Node\n')+1 : lines_callus.index('*Element, type=C3D10\n')]

# split and trim each line 
nodes_list = [[float(i) for i in item.strip().split(',')] for item in nodes_list]


# convet to data frame
nodes_df = pd.DataFrame(nodes_list,  columns=["Node", "x_cord", "y_cord", "z_cord"])


##old code
#nodes_df = pd.read_csv("node_file.txt", header = None, names=["Node", "x_cord", "y_cord", "z_cord"])


# In[ ]:


max(nodes_df.x_cord)


# In[ ]:


min(nodes_df.x_cord)


# In[ ]:


max(nodes_df.y_cord)


# In[ ]:


min(nodes_df.y_cord)


# In[ ]:


max(nodes_df.z_cord)


# In[ ]:


min(nodes_df.z_cord)


# ### 1.2 Read Element 

# In[ ]:


## Elements

# get only nodes
elements_list = lines_callus[lines_callus.index('*Element, type=C3D10\n')+1 : lines_callus.index('*Nset, nset=Set-1, generate\n')]

# split and trim each line 
elements_list = [[float(i) for i in item.strip().split(',')] for item in elements_list]


# convet to data frame
elements_df = pd.DataFrame(elements_list)
elements_df = elements_df.rename(columns={0: 'Element'})

#elements_df = pd.read_csv("element_file.txt", header = None)
#elements_df = elements_df.rename(columns={0: 'Element'})


# In[ ]:


np_nodes = nodes_df.to_numpy(copy=True)
np_elements = elements_df.to_numpy(copy=True)


# ## 2.Create Lattice

# In[ ]:


# creating lattice
# to do: automate this part



# LATTICE_X = 151
# LATTICE_Y = 151
# LATTICE_Z = 301
LATTICE_X = 101
LATTICE_Y = 161
LATTICE_Z = 101

cells =  pd.DataFrame(product(*[range(LATTICE_X),range(LATTICE_Y),range(LATTICE_Z)]), columns=['x','y','z'])
# cells[['x_mm', 'y_mm', 'z_mm']] = cells[['x', 'y', 'z']]/20
cells['x_mm']= cells['x']/20 + 8
cells['y_mm']= cells['y']/20 -1
cells['z_mm']= cells['z']/20 - 2.5
cells['age'] = 666
cells['lattice'] = 0
cells['in_elemnet'] = -1
cells['stimulus'] = -1
# cells[['x', 'y', 'z']] = cells[['x', 'y', 'z']]/100


# In[ ]:





# In[ ]:





# In[ ]:





# In[ ]:


np_cells = cells.to_numpy(copy=True)


# ### 2.1Map Lattice

# In[ ]:


# mapping lattice to abaqus part
# to do: slowest part of the code, make it faster
for i in np_elements:
    k = ConvexHull(np_nodes[np.isin(np_nodes[:,0], i[1:])][:, 1:])
    a = k.equations[:, :-1]
    b = k.equations[:, -1].reshape(-1, 1)
   
   
    x_min = min(np_nodes[np.isin(np_nodes[:,0], i[1:])][:, 1])
    y_min = min(np_nodes[np.isin(np_nodes[:,0], i[1:])][:, 2])
    z_min = min(np_nodes[np.isin(np_nodes[:,0], i[1:])][:, 3])
    x_max = max(np_nodes[np.isin(np_nodes[:,0], i[1:])][:, 1])
    y_max = max(np_nodes[np.isin(np_nodes[:,0], i[1:])][:, 2])
    z_max = max(np_nodes[np.isin(np_nodes[:,0], i[1:])][:, 3])
   
    indices_min_max = (
        (np_cells[:, 5] <= z_max) & (np_cells[:, 5] >= z_min) &
        (np_cells[:, 4] <= y_max) & (np_cells[:, 4] >= y_min) &
        (np_cells[:, 3] <= x_max) & (np_cells[:, 3] >= x_min)
    )
    cell_in_min_max = np_cells[indices_min_max]

    # Apply the condition to filter cell_in_min_max
    indices_filtered = np.all((np.dot(a, cell_in_min_max[:, 3:6].T) + b) < 0, axis=0)
    filtered_cells = cell_in_min_max[indices_filtered]
    filtered_cells[:, 8] = i[0]

    # Get the indices of np_cells that satisfy the filters
    indices_np_cells = np.where(indices_min_max)[0]

    # Replace the values in np_cells at the corresponding indices with filtered_cells
    np_cells[indices_np_cells[indices_filtered]] = filtered_cells

   
    print(i[0])


# 

# In[ ]:





# ## 3.Create FE Model Job File (Abaqus)

# In[ ]:


# to do: assumption, below code might be incorrect, could not understand how material properties are changing with cells

def prepare_simulation_file(number_elems):
    # Remove previous model file (from previous simulation)
    if os.path.exists("bone_model.txt"):
        os.remove("bone_model.txt")
    # Remove previous simulation file
    if os.path.exists("bone.inp"):
        os.remove("bone.inp")
    
    # Open files
    myfile = open("bone.inp", "w")  # initial simulation file
    mytxtfile = open("bone_model.txt", "w")  # model file to be used in later iterations
    infile = open("Job-1.inp", "r")
    
    # Read & copy job file until ELSET section
    line = infile.readline()
    while line.strip() != "** Section: Granulation_tissue":
        myfile.write(line)
        mytxtfile.write(line)
        line = infile.readline()
    
     # Write set & section descriptions (one set & section per element) into model file
        
    for elem in range(number_elems):
        elemsetname = f"Elset_{elem+1}"
        materialname = f"material_{elem+1}"
        mytxtfile.write(f"*Elset, elset={elemsetname}\n")
        mytxtfile.write(f"{elem+1}\n")
        mytxtfile.write(f"*Solid section, elset={elemsetname}, material={materialname}\n")
        mytxtfile.write("1.,\n")
    
    # Copy rest of the file
    line = infile.readline()  # *Solid Section, elset=t, material=tt
    myfile.write(line)
    line = infile.readline()  # ,
    myfile.write(line)
    line = infile.readline()  # *End Part
    myfile.write(line)
    while line.strip() != "350., 0.4":
        myfile.write(line)
        mytxtfile.write(line)
        line = infile.readline()
    mytxtfile.write(line)
    mytxtfile.close()  # close model text file
    
    while line.strip() != "** OUTPUT REQUESTS":  # copy rest to initial job file
        myfile.write(line)
        line = infile.readline()
    # Change Boundary conditions based on your case (Do not forget to change 'ELSET' to match callus part name in the model)
    # Write ending into job file (E and S output in dat file)
    myfile.write("**---------------- OUTPUT REQUESTS -------------------------------\n")
    myfile.write("** \n")
    myfile.write("*Restart, write, frequency=0\n")
    myfile.write("** \n")
    myfile.write("** FIELD OUTPUT: F-Output-1\n")
    myfile.write("**\n")
    myfile.write("*Output, field\n")
    myfile.write("*Element Output, directions=YES\n")
    myfile.write("E, S\n")
    myfile.write("**\n")
    myfile.write("**\n")
    myfile.write("*EL PRINT, FREQUENCY=100, ELSET=Callus-1.Set-1\n")
    myfile.write("EP, SP\n")
    myfile.write("**\n")
    myfile.write("**\n")
    myfile.write("** HISTORY OUTPUT: H-Output-1\n")
    myfile.write("**\n")
    myfile.write("**\n")
    myfile.write("*Output, history\n")
    myfile.write("*Energy Output\n")
    myfile.write("ALLAE,\n")
    myfile.write("*End Step\n")
    myfile.close()
    
    # Close files
    infile.close()


# ### 3.1Update FE Model Job File With New Bone Formation

# In[ ]:


def update_simulation_file(number_elems , df_cells):
    # Remove previous simulation file
    if os.path.exists("bone.inp"):
        os.remove("bone.inp")
   
    mytxtfile = open("bone_model.txt", "r")
    myfile = open("bone.inp", "w")
   
    for line in mytxtfile:
        myfile.write(line)
        
      # Biological model material properties from Table 1 in the paper 
       
    df_cells['Y'] = np.where(
        df_cells['lattice'] == 0, 0.2, np.where(
        df_cells['lattice'] == 1, 0.2, np.where(
        df_cells['lattice'] == 2, 8000, np.where(
        df_cells['lattice'] == 4, 10, np.where(
        df_cells['lattice'] == 5, 2, 0)))))
   
    df_cells['P'] = np.where(
        df_cells['lattice'] == 0, 0.167, np.where(
        df_cells['lattice'] == 1, 0.167, np.where(
        df_cells['lattice'] == 2, 0.3, np.where(
        df_cells['lattice'] == 4, 0.167, np.where(
        df_cells['lattice'] == 5, 0.167, 0)))))
   
    df_material_props = df_cells[df_cells['in_elemnet']!=-1].groupby(['in_elemnet']).agg({"Y" : "sum", "P" : "sum", "lattice" : "count"}).reset_index()
    df_material_props["Y"] = df_material_props["Y"]/df_material_props['lattice']
    df_material_props["P"] = df_material_props["P"]/df_material_props['lattice']
   
   
    for i, row in df_material_props.iterrows():
        Young_mod_elem = row.Y
        Poison_rat_elem = row.P
        materialname = f"material_{int(row.in_elemnet)}"
        myfile.write(f"*Material, name={materialname}\n")
        myfile.write("*Elastic\n")
        myfile.write(f"{Young_mod_elem}, {Poison_rat_elem}\n")
        
    all_numbers = list(range(1, number_elems))
   
    missing_elements = [x for x in all_numbers if x not in list(df_material_props['in_elemnet'])]
   
    for i in missing_elements:
        Young_mod_elem = 0.2
        Poison_rat_elem = 0.167
        materialname = f"material_{int(i)}"
        myfile.write(f"*Material, name={materialname}\n")
        myfile.write("*Elastic\n")
        myfile.write(f"{Young_mod_elem}, {Poison_rat_elem}\n")
       
    # Change Boundary conditions based on your case (Do not forget to change 'ELSET' to match callus part name in the model)
    myfile.write("**\n")
    myfile.write("** ----------------------------------------------------------------\n")
    myfile.write("**\n")
    myfile.write("** STEP: Step-1\n")
    myfile.write("** \n")
    myfile.write("*Step, name=Step-1, nlgeom=NO\n")
    myfile.write("*Static, stabilize=0.0002, allsdtol=0.05, continue=NO\n")
    myfile.write("0.1, 1., 1e-05, 1.,\n")
    myfile.write("** \n")
    myfile.write("**-------------- BOUNDARY CONDITIONS   ------------------------------------------\n")
    myfile.write("**\n")
    myfile.write("**\n")
    myfile.write("** Name: BC-1 Type: Displacement/Rotation\n")
    myfile.write("*Boundary\n")
    myfile.write("Set-42, 1, 1\n")
    myfile.write("Set-42, 2, 2\n")
    myfile.write("Set-42, 3, 3\n")
    myfile.write("Set-42, 4, 4\n")
    myfile.write("Set-42, 5, 5\n")
    myfile.write("Set-42, 6, 6\n")
    myfile.write("**\n")
    myfile.write("**--------------  Loads  ------------------------------------------\n")
    myfile.write("**\n")
    myfile.write("**\n")
    myfile.write("** Name: Load-1   Type: Pressure\n")
    myfile.write("*Dsload\n")
    myfile.write("Bone_up_load, P, 17.7\n")
    myfile.write("** Name: Load-3   Type: Concentrated force\n")
    myfile.write("*Cload\n")
    myfile.write("m_Set-43, 1, -5.7\n")
    myfile.write("**\n")
    myfile.write("**---------------- OUTPUT REQUESTS -------------------------------\n")
    myfile.write("** \n")
    myfile.write("*Restart, write, frequency=0\n")
    myfile.write("** \n")
    myfile.write("** FIELD OUTPUT: F-Output-1\n")
    myfile.write("**\n")
    myfile.write("*Output, field\n")
    myfile.write("*Element Output, directions=YES\n")
    myfile.write("E, S\n")
    myfile.write("**\n")
    myfile.write("**\n")
    myfile.write("*EL PRINT, FREQUENCY=100, ELSET=Callus-1.Set-1\n")
    myfile.write("EP, SP\n")
    myfile.write("**\n")
    myfile.write("**\n")
    myfile.write("** HISTORY OUTPUT: H-Output-1\n")
    myfile.write("**\n")
    myfile.write("**\n")
    myfile.write("*Output, history\n")
    myfile.write("*Energy Output\n")
    myfile.write("ALLAE,\n")
    myfile.write("*End Step\n")
    myfile.close()
   
    mytxtfile.close()  # close model text file


# In[ ]:





# In[ ]:





# # 4.Functions

# ### 4.1Read Stimulus

# In[ ]:


def calculate_simulus():
    
    # read dat file
    with open('bone.dat') as f:
        lines = f.readlines()
        
    # to verify: below logic always holds true
    lines_sim_start = lines.index('    ELEMENT  PT FOOT-       EP1         EP2         EP3         SP1         SP2         SP3         \n') + 3
    lines_sim_end = lines.index('          THE ANALYSIS HAS BEEN COMPLETED\n') - 8
    
    lines_sim = lines[lines_sim_start:lines_sim_end]
    
    # split and trim each line 
    lines_sim = [[float(i) for i in item.strip().split()] for item in lines_sim]
    
    # convet to data frame
    sim_val_df = pd.DataFrame(lines_sim,  columns=["ELEMENT", "PT", "EP1", "EP2", "EP3", "SP1", "SP2", "SP3"])
    
    # add calulated column to data frame
    # to do : add logic in coment
    sim_val_df["HP_read"] = - sim_val_df[["SP1", "SP2", "SP3"]].sum(axis = 1) / 3
    sim_proc_df = (sim_val_df.groupby("ELEMENT")[["EP1", "HP_read"]].sum() / 4).reset_index()
    
    cond1 = np.logical_and(abs(sim_proc_df['HP_read']) < 0.0016, abs(sim_proc_df['EP1']) < 0.0004)
    cond2 = np.logical_and(abs(sim_proc_df['HP_read']) < 0.15, abs(sim_proc_df['EP1']) < 0.05)
    cond3 = np.logical_and(sim_proc_df['HP_read'] < -0.15, abs(sim_proc_df['EP1']) < 0.15)

    sim_proc_df['cal_stimulus'] = np.where(cond1, 1, np.where(cond2, 2, np.where(cond3, 4, 5)))
    
    return sim_proc_df
    


# ## 4.2.Cell Proliferation

# In[ ]:


def func_poli(f_pol, df_cells, n_lattice):
    # mitosis

    # ii = 0
    for (x, y, z), row in f_pol.iterrows():
    #     ii+=1
    #     print(ii)
        # define the dictionary of possibilities and their corresponding coordinate changes
        possibilities = {
        0: (1, 0, 0),
        1: (-1, 0, 0),
        2: (0, 1, 0),
        3: (0, -1, 0),
        4: (0, 0, 1),
        5: (0, 0, -1)
    }
        mitosis = 0
        while not mitosis and len(possibilities)>0:
            r = random.choices(list(possibilities.keys()))[0]
    #         print(r)
            dx, dy, dz = possibilities[r]
            if ((0 <= x + dx < LATTICE_X) and (0 <= y + dy < LATTICE_Y) and (0 <= z + dz < LATTICE_Z) and
                (df_cells.at[(x+dx, y+dy, z+dz), 'lattice'] == 0).all() and (df_cells.at[(x+dx, y+dy, z+dz), 'in_elemnet'] != -1).all()):
                # the cell can proliferate
                df_cells.at[(x+dx, y+dy, z+dz), 'lattice'] = 100
                df_cells.at[(x, y, z), 'lattice'] = 100
                df_cells.at[(x, y, z), 'age'] = 0
                mitosis = 1

            else:
                possibilities.pop(r)
                
                
    df_cells.loc[df_cells['lattice'] == 100, 'age'] = 1
    df_cells.loc[df_cells['lattice'] == 100, 'lattice'] = n_lattice
    
    return df_cells


# ## 4.3.Cell Migration

# In[ ]:


def func_mig(f_mig, df_cells, number_jumps):

    # migration
#     ii = 0
    for i in range(number_jumps):
#         ii+=1
#         print(ii)
        df_cells.loc[df_cells['lattice'] == 100, 'lattice'] = 1
        f_mig = df_cells[(df_cells['in_elemnet']!=-1) & (df_cells['lattice'] == 1)].sample(frac=1)
        for (x, y, z), row in f_mig.iterrows():
            # define the dictionary of possibilities and their corresponding coordinate changes
            possibilities = {
            0: (1, 0, 0),
            1: (-1, 0, 0),
            2: (0, 1, 0),
            3: (0, -1, 0),
            4: (0, 0, 1),
            5: (0, 0, -1)
        }
            mig = 0
            while not mig and len(possibilities)>0:
                r = random.choices(list(possibilities.keys()))[0]
        #         print(r)
                dx, dy, dz = possibilities[r]
                if ((0 <= x + dx < LATTICE_X) and (0 <= y + dy < LATTICE_Y) and (0 <= z + dz < LATTICE_Z) and
                    (df_cells.at[(x+dx, y+dy, z+dz), 'lattice'] == 0).all()):
                    # the cell can proliferate
                    df_cells.at[(x+dx, y+dy, z+dz), 'lattice'] = 100
                    df_cells.at[(x+dx, y+dy, z+dz), 'age'] = df_cells.at[(x, y, z), 'age']
                    df_cells.at[(x, y, z), 'lattice'] = 0
                    df_cells.at[(x, y, z), 'age'] = 0
                    mig = 1

                else:
                    possibilities.pop(r)
        df_cells.loc[(df_cells['in_elemnet'] == -1), 'age'] = 666
    return df_cells


# ## 4.4.Cell Differentiation & Apoptosis

# In[ ]:


def func_diff(df_cells, mature_age, it):

    f_mature = df_cells_new[(df_cells_new['age'] >= mature_age)&(df_cells_new['age'] != 666)]
    
    # for MSC
    if it <= ACTIVITY_MAX:
        f_mature_msc = f_mature[f_mature['lattice'] == 1]
        n_msc = f_mature_msc.shape[0]
        f_mature_msc_to_diff = f_mature_msc.sample(n = math.floor(n_msc*0.3))
        
    else:
        f_mature_msc = f_mature[f_mature['lattice'] == 1]
        n_msc = f_mature_msc.shape[0]
        f_mature_msc_to_diff = f_mature_msc.sample(n = math.floor(n_msc*0.06))
        
    for (x, y, z), row in f_mature_msc_to_diff.iterrows():
        df_cells.at[(x, y, z), 'lattice'] = row.cal_stimulus
        df_cells.at[(x, y, z), 'age'] = 1
    
    # Apoptosis
    
    # fibroblasts
    m_cho = df_cells[(df_cells['in_elemnet']!=-1) & (df_cells['lattice'] == 5) & (df_cells['cal_stimulus'].isin([1, 2, 4]))]
    n_m_cho = m_cho.shape[0]
    m_cho_to_remove = m_cho.sample(n = math.floor(n_m_cho*0.05))
    
    for (x, y, z), row in m_cho_to_remove.iterrows():
        df_cells.at[(x, y, z), 'lattice'] = 0
        df_cells.at[(x, y, z), 'age'] = 0
    
    
    # chondrocytes
    m_cho = df_cells[(df_cells['in_elemnet']!=-1) & (df_cells['lattice'] == 4) & (df_cells['cal_stimulus'].isin([1, 2, 5]))]
    n_m_cho = m_cho.shape[0]
    m_cho_to_remove = m_cho.sample(n = math.floor(n_m_cho*0.1))
    
    for (x, y, z), row in m_cho_to_remove.iterrows():
        df_cells.at[(x, y, z), 'lattice'] = 0
        df_cells.at[(x, y, z), 'age'] = 0
        
        
    # mature_osteoblasts
    m_ost = df_cells[(df_cells['in_elemnet']!=-1) & (df_cells['lattice'] == 2) & (df_cells['cal_stimulus'] == 1)]
    n_m_ost = m_ost.shape[0]
    m_ost_to_remove = m_ost.sample(n = math.floor(n_m_ost*0.16))
    
    for (x, y, z), row in m_ost_to_remove.iterrows():
        df_cells.at[(x, y, z), 'lattice'] = 0
        df_cells.at[(x, y, z), 'age'] = 0
    
    
    return df_cells


# In[ ]:





# In[ ]:





# In[ ]:





# In[ ]:


df_cells = pd.DataFrame(np_cells, columns =['x', 'y', 'z', 'x_mm', 'y_mm', 'z_mm', 'age', 'lattice', 'in_elemnet', 'stimulus'])


# In[ ]:


df_cells.reset_index(inplace=True)
df_cells.set_index(['x', 'y', 'z'], inplace=True)


# # 5.Biological Model Boundary Conditions

# In[ ]:





# In[ ]:


# assigning age = 0
df_cells.loc[df_cells['in_elemnet']!=-1, 'age'] = 0


# In[ ]:


# seeding MSC from top and bottom
df_cells.loc[(df_cells['in_elemnet']!=-1) & ((df_cells['y_mm']<(0.5)) | (df_cells['y_mm'] > (5.5))), 'lattice'] = 1


# In[ ]:


df_cells.loc[(df_cells['in_elemnet']!=-1) & ((df_cells['y_mm']<(0.5)) | (df_cells['y_mm'] > (5.5))), 'age'] = 1


# In[ ]:


# df_cells[df_cells['lattice'] == 1]


# In[ ]:





# In[ ]:


input_poli = pd.DataFrame({'lattice':[1, 5, 4, 2],'poli_ratio':[0.6, 0.55, 0.2, 0.3]}, columns = ['lattice', 'poli_ratio'])


# In[ ]:


input_poli['poli_ratio_2'] = input_poli['poli_ratio']/5


# In[ ]:


input_poli


# In[ ]:


input_diff = pd.DataFrame({'lattice':[1, 5, 4, 2],'apoptosis_rate':[0.05, 0.05, 0.1, 0.16], 'diff':[0.3, 0, 0, 0]
                           , 'diff_2':[0.06, 0, 0, 0]}, columns = ['lattice', 'apoptosis_rate', 'diff', 'diff_2'])


# In[ ]:


input_diff


# # 6.Automation Set up (Healing Period, Latency Period, and Running FE Model).

# In[ ]:


mature_cell_age = 6


# In[ ]:


iteration_days = 112


# In[ ]:


ACTIVITY_MAX = 7


# In[ ]:


for ii in range(iteration_days):
    print('Day '+str(ii+1))
    if (ii == 0):
        
        prepare_simulation_file(len(np_elements))
        print('run abaqus')
        subprocess.call(["bone_run_abaqus.bat"])
        sim_final_df = calculate_simulus()
        df_cells_new = df_cells.merge(sim_final_df[['ELEMENT', 'cal_stimulus']]
                                      , left_on = 'in_elemnet', right_on = 'ELEMENT', how = 'left' , right_index=True)
    
    if ((ii+1)%3 == 0):
        update_simulation_file(len(np_elements), df_cells_new)
        print('run abaqus')
        subprocess.call(["bone_run_abaqus.bat"])
        sim_final_df = calculate_simulus()
        df_cells_new = df_cells_new[['index', 'x_mm', 'y_mm', 'z_mm', 'age', 'lattice', 'in_elemnet']].merge(sim_final_df[['ELEMENT', 'cal_stimulus']]
                                      , left_on = 'in_elemnet', right_on = 'ELEMENT', how = 'left' , right_index=True)
    
    # dif
    if ii > 5:
        # to do
        print("Cell differentiation")
        df_cells_new = func_diff(df_cells_new, mature_cell_age, ii)
        
    # pol
    print("Cell proliferation")
    count_cell_type = df_cells_new[df_cells_new['in_elemnet']!=-1].groupby(['lattice'])['lattice'].count().reset_index(name='cell_count')
    count_cell_type = count_cell_type.merge(input_poli, on=['lattice'], how='left')
    
    count_cell_type['cells_to_prol'] = (count_cell_type['cell_count'] * count_cell_type['poli_ratio'])
    count_cell_type['cells_to_prol_2'] = (count_cell_type['cell_count'] * count_cell_type['poli_ratio_2'])
    
    for ind , row in count_cell_type.iterrows():
        if (row.lattice != 0):
            if ii < ACTIVITY_MAX:
                f_pol_a = df_cells_new[(df_cells_new['lattice'] == row.lattice)&(df_cells_new['in_elemnet']!=-1)
                                    ].sample(n = math.floor(row.cells_to_prol))
            else:
                f_pol_a = df_cells_new[(df_cells_new['lattice'] == row.lattice)&(df_cells_new['in_elemnet']!=-1)
                                    ].sample(n = math.floor(row.cells_to_prol_2))

            df_cells_new = func_poli(f_pol_a, df_cells_new, row.lattice)
        
    # mig
    print("Cell migration")
    f_mig_a = df_cells_new[(df_cells_new['in_elemnet']!=-1) & (df_cells_new['lattice'] == 1)].sample(frac=1)
    # velocity = 30 micrometers/hour (30*24/(1/x)); x=the number of cells in lattice creation code above.
    ## pending task change belove 7 to auto adjust speed
    n_jumps = 14
    
    df_cells_new = func_mig(f_mig_a, df_cells_new, n_jumps)
    df_cells_new.loc[df_cells_new['lattice'] == 100, 'lattice'] = 1
    
    print("Write results file")
    result_file_name = 'raw_cells_output_iteration_' + str(ii)+ '.csv'
    df_cells_new.to_csv(result_file_name)
    
    
    print("increase age")
    df_cells_new.loc[(df_cells_new['age']>0)&(df_cells_new['age']!=666), 'age'] += 1
        
    
        
        

    
    
    


# In[ ]:


import plotly.graph_objs as go
import numpy as np

# Create data array
data = df_cells_new[df_cells_new['lattice'] == 1][['x_mm', 'y_mm', 'z_mm']].to_numpy()

# Create trace for scatter plot
trace = go.Scatter3d(
    x=data[:, 0],
    y=data[:, 1],
    z=data[:, 2],
    mode='markers',
    marker=dict(
        size=5,
        color=data[:, 2], 
        colorscale='Viridis', 
        opacity=0.8
    )
)

# Create layout for plot
layout = go.Layout(
    margin=dict(l=0, r=0, b=0, t=0),
    scene=dict(
        camera=dict(
            up=dict(x=0, y=0, z=1),
            center=dict(x=0, y=0, z=0),
            eye=dict(x=1.5, y=1.5, z=1.5)
        ),
        xaxis=dict(title='X'),
        yaxis=dict(title='Y'),
        zaxis=dict(title='Z')
    )
)

# Create Figure object and plot it
fig = go.Figure(data=[trace], layout=layout)
fig.show()


# In[ ]:





# In[ ]:


import plotly.graph_objs as go
import numpy as np

# Create data array
data = df_cells_new[df_cells_new['lattice'] == 2][['x_mm', 'y_mm', 'z_mm']].to_numpy()

# Create trace for scatter plot
trace = go.Scatter3d(
    x=data[:, 0],
    y=data[:, 1],
    z=data[:, 2],
    mode='markers',
    marker=dict(
        size=5,
        color=data[:, 2], 
        colorscale='gray', 
        opacity=0.8
    )
)

# Create layout for plot
layout = go.Layout(
    margin=dict(l=0, r=0, b=0, t=0),
    scene=dict(
        camera=dict(
            up=dict(x=0, y=0, z=1),
            center=dict(x=0, y=0, z=0),
            eye=dict(x=1.5, y=1.5, z=1.5)
        ),
        xaxis=dict(title='X'),
        yaxis=dict(title='Y'),
        zaxis=dict(title='Z')
    )
)

# Create Figure object and plot it
fig = go.Figure(data=[trace], layout=layout)
fig.show()

